/*
 * BearLibFOV.cpp
 *
 *  Created on: Jun 6, 2013
 *      Author: cfyz
 */

#define BEARLIBFOV_LIBRARY
#include <BearLibFOV.h>

#include <BearLibMap.h>
#include <math.h>
#include <stdio.h>
#include <string.h>
#include <algorithm>
#include <vector>
#include <cstdint>

struct shadow_t
{
    double start;
    double end;

    shadow_t(): start(0), end(0)
    { }

    shadow_t(double start, double end): start(start), end(end)
    { }
};

struct fov_context
{
	int map_id;
	int map_width;
	int map_height;
	int opacity_layer;
	fov_callback callback;
	void* opaque;
	std::vector<shadow_t> shadows;
	std::vector<std::uint8_t> visible;
};

inline bool CellOpaque(fov_context& context, int x, int y)
{
    if (x < 0 || y < 0 || x >= context.map_width || y >= context.map_height ) return false;
    return map_get(context.map_id, x, y, context.opacity_layer);
}

inline bool CellVisible(fov_context& context, int x, int y)
{
	if (x < 0 || y < 0 || x >= context.map_width || y >= context.map_height) return false;
	return context.visible[y*context.map_width+x];
}

inline void SetVisibility(fov_context& context, int x, int y)
{
	if (x < 0 || y < 0 || x >= context.map_width || y >= context.map_height) return;
	if (!context.visible[y*context.map_width+x])
	{
		context.visible[y*context.map_width+x] = 1;
		context.callback(context.map_id, x, y, context.opaque);
	}
}

void CalcFOVQuadrant3(int px, int py, int radius, int xy, int dx, int dy, fov_context& context)
{
    int iteration = 1; // iteration step (distance from (px,py) by main direction)

    int shadows_before = 0;
	double min_angle = 0.0f;

	int x, y, dv, dw, pv, pw, maxv;
    int tx, ty;

    // xy: 0 = vertical (y) edge, 1 = horisontal (x) edge
    int& v = xy ? x : y;
    int& w = xy ? y : x;

    if ( xy == 0 )
    {
        // Vertical edge
        pv = py; pw = px;
        dv = dy; dw = dx;
        tx = 0; ty = dy;
        maxv = context.map_height;
    }
    else
    {
        // Horisontal edge
        pv = px; pw = py;
        dv = dx; dw = dy;
        tx = dx; ty = 0;
        maxv = context.map_width;
    }

    // Reset shadow counter
    context.shadows.resize(0);

    // Do while there are unblocked slopes left and the algo is within the map's boundaries
    // Scan progressive lines/columns from the PC outwards
    v = pv + dv; // The outer slope's coordinates (first processed line)

    bool done = false;
    if (v < 0 || v >= maxv) done = true;

	while (!done)
    {
        // Process cells in the line
		double slopesPerCell = 1.0 / (double)(iteration + 1);
		double halfSlopes = slopesPerCell * 0.5;
		int processedCell = (int)floor(min_angle / slopesPerCell);

        int minw = std::max(0, pw-iteration);
        int maxw = std::min((xy? context.map_height: context.map_width)-1, pw+iteration);

        done = true; // If there will be any visible cells in current line, this flag will be cleared back to false

        for (w = pw + (processedCell * dw); w >= minw && w <= maxw; w += dw)
        {
            bool visible = true;

            double startSlope = (double)processedCell*slopesPerCell;
            double centreSlope = startSlope+halfSlopes;
            double endSlope = startSlope+slopesPerCell;

            if (shadows_before > 0 && !CellVisible(context, x, y))
            {
            	for (auto& shadow: context.shadows)
				{
					if (!CellOpaque(context, x, y)) // Can see throug this cell?
					{
						if (centreSlope > shadow.start && centreSlope < shadow.end)
						{
							visible = false;
						}
					}
					else
					{
						if (startSlope >= shadow.start && endSlope <= shadow.end)
						{
							visible = false;
						}
					}

					// If cell is still marked as visible it can be a glitch
					// To make sure, check if two nearby cells in previous line are blocked from sight or not
					bool c1blocked = !CellVisible(context, x-tx, y-ty) || CellOpaque(context, x-tx, y-ty);
					bool c2blocked = !CellVisible(context, x-dx, y-dy) || CellOpaque(context, x-dx, y-dy);

					// If there are no close and visible cells in previous line, then it's a glitch
					if (visible && c1blocked && c2blocked) visible = false;
				}
            }

            if (visible)
            {
                if ((x-px)*(x-px) + (y-py)*(y-py) <= radius*radius)
                {
                	SetVisibility(context, x, y);
                }

                done = false; // Visible cell means there can be more visible cells further, so we're not done

                // If the cell is opaque, block the adjacent slopes
                if (CellOpaque(context, x, y))
                {
                    if (min_angle >= startSlope)
                    {
                        min_angle = endSlope;
                    }
                    else
                    {
                        // TODO: Shadow optimization is not implemented yet
                        // TODO: CombineShadows(shadow_t(startSlope, endSlope));
                    	context.shadows.emplace_back(startSlope, endSlope);
                    }
                }
            }

            processedCell++;
        }

        if (iteration == radius) done = true;
        iteration++;

        shadows_before = context.shadows.size();

        v += dv;

        if (v < 0 || v >= maxv) done = true;
		if (min_angle >= 1.0f) done = true;
    }
}

int fov_calc_cb(int map_id, int opacity_layer, int x, int y, int radius, fov_callback callback, void* opaque)
{
	fov_context context;
	context.map_id = map_id;
	context.map_width = map_width(map_id);
	context.map_height = map_height(map_id);
	context.opacity_layer = opacity_layer;
	context.callback = callback;
	context.opaque = opaque;

	// TODO: allocation cost may be reduced via caching
	context.visible.resize(context.map_width*context.map_height);
	memset(context.visible.data(), 0, context.visible.size());

	SetVisibility(context, x, y);

	const int vectors[4][2] =
	{
		{ +1, +1 },
		{ +1, -1 },
		{ -1, +1 },
		{ -1, -1 }
	};

	for ( int i=0; i<4; i++ )
	{
		CalcFOVQuadrant3(x, y, radius, 0, vectors[i][0], vectors[i][1], context);
		CalcFOVQuadrant3(x, y, radius, 1, vectors[i][0], vectors[i][1], context);
	}

	return 0;
}

static void fov_calc_callback(int map_id, int x, int y, void* opaque)
{
	int visibility_layer = *(int*)opaque;
	map_set(map_id, x, y, visibility_layer, 1);
}

int fov_calc(int map_id, int opacity_layer, int visibility_layer, int x, int y, int radius)
{
	return fov_calc_cb(map_id, opacity_layer, x, y, radius, fov_calc_callback, &visibility_layer);
}
