/*
* BearLibMap
* Copyright (C) 2013 Cfyz
*
* Permission is hereby granted, free of charge, to any person obtaining a copy
* of this software and associated documentation files (the "Software"), to deal
* in the Software without restriction, including without limitation the rights
* to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
* of the Software, and to permit persons to whom the Software is furnished to do
* so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in all
* copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
* FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
* COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
* IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
* CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#define BEARLIBMAP_LIBRARY

#include <string>
#include <vector>
#include <cstdint>
#include <cstddef>
#include <cstring>
#include <utility>
#include <limits>
#include <stdexcept>
#include "BearLibMap.h"
#include "Utility.hpp"

namespace BearLibMap
{
	enum class Type
	{
		Flag, Integer, Float
	};

	struct Level
	{
		Level(Type type, size_t field_offset, size_t data_offset);

		Type type;
		size_t field_offset;
		size_t data_offset;
	};

	struct Description
	{
		Description(const std::string& description);

		std::vector<Level> levels; // Fields
		size_t stride; // Size of one cell in bytes
	};

	class Map
	{
	public:
		Map();
		Map(int width, int height, const std::string& description);
		Map(const Map& from);
		Map(Map&& from);
		~Map();
		Map& operator=(Map another);
		void Swap(Map& another);
		bool IsEmpty() const;
		int GetWidth() const;
		int GetHeight() const;
		int GetLevelCount() const;
		void Clear();
		void Clear(int level);
		template<typename T> void Set(int x, int y, int level, T value);
		template<typename T> T Get(int x, int y, int level) const;
	private:
		int m_width;
		int m_height;
		Description m_description;
		uint8_t* m_data;
	};

	static std::vector<Map> maps;
}


#include <iostream>

namespace BearLibMap
{
	Level::Level(Type type, size_t field_offset, size_t data_offset):
		type(type),
		field_offset(field_offset),
		data_offset(data_offset)
	{ }

	Description::Description(const std::string& description)
	{
		size_t flags_count = 0;
		size_t flags_field = 0;
		size_t fields = 0;

		for (auto& type: BearLib::split(description, ','))
		{
			if (type == "flag" && flags_count < 32)
			{
				if (flags_count == 0) flags_field = fields++; // First flag
				levels.emplace_back(Type::Flag, flags_field*4, flags_count++);
			}
			else if (type == "integer")
			{
				levels.emplace_back(Type::Integer, fields++*4, 0);
			}
			else if (type == "float")
			{
				levels.emplace_back(Type::Float, fields++*4, 0);
			}
		}

		stride = fields*4;
	}

	template<typename T> void SetField(uint8_t* p, Level& level, T value);

	template<> void SetField<bool>(uint8_t* p, Level& level, bool value)
	{
		uint32_t& flags = *(uint32_t*)p;
		if (value) flags |= (1 << level.data_offset); else flags &= ~(1 << level.data_offset);
	}

	template<> void SetField<int>(uint8_t* p, Level& level, int value)
	{
		*(int32_t*)p = value;
	}

	template<> void SetField<float>(uint8_t* p, Level& level, float value)
	{
		*(float*)p = value;
	}

	template<typename T> T GetField(uint8_t* p, const Level& level);

	template<> bool GetField<bool>(uint8_t* p, const Level& level)
	{
		uint32_t& flags = *(uint32_t*)p;
		return flags & (1 << level.data_offset);
	}

	template<> int GetField<int>(uint8_t* p, const Level& level)
	{
		return *(int32_t*)p;
	}

	template<> float GetField<float>(uint8_t* p, const Level& level)
	{
		return *(float*)p;
	}

	Map::Map():
		m_width(0),
		m_height(0),
		m_description(""),
		m_data(nullptr)
	{ }

	Map::Map(int width, int height, const std::string& description):
		m_width(width),
		m_height(height),
		m_description(description),
		m_data(nullptr)
	{
		if (m_description.levels.size() == 0) throw std::runtime_error("Empty description");
		size_t size = m_width*m_height*m_description.stride;
		m_data = new uint8_t[size];
	}

	Map::Map(const Map& from):
		m_width(from.m_width),
		m_height(from.m_height),
		m_description(from.m_description),
		m_data(nullptr)
	{
		size_t size = m_width*m_height*m_description.stride;
		m_data = new uint8_t[size];
		std::memcpy(m_data, from.m_data, size);
	}

	Map::Map(Map&& from):
		m_width(0),
		m_height(0),
		m_description(""),
		m_data(nullptr)
	{
		Swap(from);
	}

	Map::~Map()
	{
		if (m_data != nullptr) delete[] m_data;
	}

	Map& Map::operator=(Map another)
	{
		Swap(another);
		return *this;
	}

	void Map::Swap(Map& another)
	{
		std::swap(m_width, another.m_width);
		std::swap(m_height, another.m_height);
		std::swap(m_description, another.m_description);
		std::swap(m_data, another.m_data);
	}

	bool Map::IsEmpty() const
	{
		return m_data == nullptr;
	}

	int Map::GetWidth() const
	{
		return m_width;
	}

	int Map::GetHeight() const
	{
		return m_height;
	}

	int Map::GetLevelCount() const
	{
		return m_description.levels.size();
	}

	void Map::Clear()
	{
		if (m_data)
		{
			memset(m_data, 0, m_width*m_height*m_description.stride);
		}
	}

	void Map::Clear(int layer)
	{
		if (layer < 0 || layer >= (int)m_description.levels.size())
		{
			return;
		}

		Level& l = m_description.levels[layer];

		for (int y=0; y<m_height; y++)
		{
			for (int x=0; x<m_width; x++)
			{
				uint8_t* data = m_data + (y*m_width+x)*m_description.stride + l.field_offset;

				switch (l.type)
				{
				case Type::Flag:
					SetField<bool>(data, l, false);
					break;
				case Type::Integer:
					SetField<int>(data, l, 0);
					break;
				case Type::Float:
					SetField<float>(data, l, 0.0f);
					break;
				}
			}
		}
	}

	template<typename T> T Map::Get(int x, int y, int layer) const
	{
		if (x < 0 || x >= m_width || y < 0 || y >= m_height || layer < 0 || layer >= m_description.levels.size())
		{
			return T();
		}

		const Level& l = m_description.levels[layer];
		uint8_t* data = m_data + (y*m_width + x)*m_description.stride + l.field_offset;

		switch (l.type)
		{
		case Type::Flag:
			return GetField<bool>(data, l);
		default:
			return GetField<T>(data, l);
		}
	}

	template<typename T> void Map::Set(int x, int y, int layer, T value)
	{
		if (x < 0 || x >= m_width || y < 0 || y >= m_height || layer < 0 || layer >= m_description.levels.size())
		{
			return;
		}

		Level& l = m_description.levels[layer];
		uint8_t* data = m_data + y*m_width*m_description.stride + x*m_description.stride + l.field_offset;

		switch (l.type)
		{
		case Type::Flag:
			return SetField<bool>(data, l, value);
		default:
			return SetField<T>(data, l, value);
		}
	}
}

static int map_acquire()
{
	int id = 0;

	while (id < (int)BearLibMap::maps.size() && !BearLibMap::maps[id].IsEmpty())
	{
		id += 1;
	}

	if (id >= (int)BearLibMap::maps.size())
	{
		BearLibMap::maps.push_back(BearLibMap::Map());
	}

	return id;
}

static inline bool check_map_id(int map_id)
{
	return map_id >= 0 && map_id < (int)BearLibMap::maps.size();
}

int map_alloc(int width, int height, const char* description)
{
	static const int max_size = std::numeric_limits<uint16_t>::max();
	if (width <= 0 || height <= 0 || width > max_size || height > max_size) return -1;

	try
	{
		int id = map_acquire();
		BearLibMap::maps[id] = BearLibMap::Map(width, height, description);
		return id;
	}
	catch(...)
	{
		return -1;
	}
}

void map_free(int map_id)
{
	if (!check_map_id(map_id)) return;
	BearLibMap::maps[map_id] = BearLibMap::Map();
}

int map_copy(int map_id)
{
	if (!check_map_id(map_id)) return -1;
	int copy_id = map_acquire();
	BearLibMap::maps[copy_id] = BearLibMap::maps[map_id];
	return copy_id;
}

int map_assign(int dst_map_id, int src_map_id)
{
	if (!check_map_id(dst_map_id) || !check_map_id(src_map_id)) return -1;
	BearLibMap::maps[dst_map_id] = BearLibMap::maps[src_map_id];
	return dst_map_id;
}

void map_clear(int map_id)
{
	if (!check_map_id(map_id)) return;
	BearLibMap::maps[map_id].Clear();
}

void map_clear_layer(int map_id, int layer)
{
	if (!check_map_id(map_id)) return;
	BearLibMap::maps[map_id].Clear(layer);
}

void map_set(int map_id, int x, int y, int layer, int value)
{
	if (!check_map_id(map_id)) return;
	BearLibMap::maps[map_id].Set<int>(x, y, layer, value);
}

void map_setf(int map_id, int x, int y, int layer, float value)
{
	if (!check_map_id(map_id)) return;
	BearLibMap::maps[map_id].Set<float>(x, y, layer, value);
}

int map_get(int map_id, int x, int y, int layer)
{
	if (!check_map_id(map_id)) return 0;
	return BearLibMap::maps[map_id].Get<int>(x, y, layer);
}

float map_getf(int map_id, int x, int y, int layer)
{
	if (!check_map_id(map_id)) return 0;
	return BearLibMap::maps[map_id].Get<float>(x, y, layer);
}

int map_width(int map_id)
{
	if (!check_map_id(map_id)) return 0;
	return BearLibMap::maps[map_id].GetWidth();
}

int map_height(int map_id)
{
	if (!check_map_id(map_id)) return 0;
	return BearLibMap::maps[map_id].GetHeight();
}
