#include "BearLibMap.h"
#include "BearLibFOV.h"
#include "BearLibTerminal.h"
#include <string>
#include <vector>
#include <list>
#include <fstream>
#include <map>
#include <sstream>
#include <iostream>
#include <math.h>

// Map layers:
const int kOpacityLayer = 0;
const int kFOVLayer = 1;
const int kColorLayer = 2;

// Platform-specific
void ErrorMessage(const std::string& message);

struct Color3i
{
	Color3i():
		r(0),
		g(0),
		b(0)
	{ }
	
	Color3i(color_t color):
		r(color_red(color)),
		g(color_green(color)),
		b(color_blue(color))
	{ }
	
	Color3i& operator+= (const Color3i& other)
	{
		r += other.r;
		g += other.g;
		b += other.b;
		return *this;
	}
	
	Color3i operator+ (const Color3i& other)
	{
		Color3i result(*this);
		result += other;
		return result;
	}
	
	operator color_t()
	{
		return color_from_argb(255, r, g, b);
	}
	
	int r, g, b;
};

struct Light
{
	Light():
		x(0),
		y(0),
		radius(5),
		enabled(true)
	{ }
	
	Color3i color;
	int x, y;
	int radius;
	bool enabled;
};

int map_id = -1;
int player_sight_radius = 15;
std::map<int, Light> lights;
Light& player = lights[0];
bool advanced = false;

std::map<int, std::pair<int, int>> kMovement =
{
	{VK_LEFT,    {-1,  0}},
	{VK_UP,      { 0, -1}},
	{VK_RIGHT,   {+1,  0}},
	{VK_DOWN,    { 0, +1}},
	{VK_NUMPAD1, {-1, +1}},
	{VK_NUMPAD2, { 0, +1}},
	{VK_NUMPAD3, {+1, +1}},
	{VK_NUMPAD4, {-1,  0}},
	{VK_NUMPAD6, {+1,  0}},
	{VK_NUMPAD7, {-1, -1}},
	{VK_NUMPAD8, { 0, -1}},
	{VK_NUMPAD9, {+1, -1}},
};

bool LoadMap()
{
	std::ifstream file("map.txt", std::ios_base::in);
	if (!file.good()) return false;
	
	std::string line;
	std::vector<std::string> map_lines;
	int width = 0, height = 0;
	while (std::getline(file, line) && !line.empty())
	{
		map_lines.push_back(line);
		width = std::max<int>(width, line.length());
		height += 1;
	}
	if (map_lines.empty()) return false;
	
	map_id = map_alloc(width, height, "integer, flag, integer"); // Opacity, FOV, color
	for (size_t y=0; y<map_lines.size(); y++)
	{
		std::string& row = map_lines[y];
		for (size_t x=0; x<row.length(); x++)
		{
			char c = row[x];
			if (c == '#')
			{
				map_set(map_id, x, y, kOpacityLayer, 1);
			}
			else if (c == '@' || (c >= '1' && c <= '9'))
			{
				auto& light = lights[c == '@'? 0: (c - '0')];
				light.x = x;
				light.y = y;
			}
		}
	}
	
	while (std::getline(file, line))
	{
		if (line.empty() || line.find("//") == 0) continue;
	
		// <id>: <radius> <color-name>
		std::istringstream ss(line);
		int id, radius;
		ss >> id;
		ss.ignore(1);
		ss >> radius;
		ss.ignore(1);
		std::string color;
		std::getline(ss, color);
		
		if (id < 0 || id > 9) continue;
		Light& light = lights[id];
		light.radius = radius;
		light.color = color_from_name(color);
		light.enabled = true;
	}
	
	return true;
}

void LightFOVCallback(int map_id, int x, int y, void* opaque)
{
	Light* light = (Light*)opaque;
	float dx = light->x-x;
	float dy = light->y-y;
	float dist2 = dx*dx+dy*dy;
	int radius2 = light->radius*light->radius;
	float factor = 1.0f - dist2/radius2;
	if (factor < 0.0f) factor = 0.0f;
	
	Color3i color = (color_t)map_get(map_id, x, y, kColorLayer);	
	color.r += (int)(light->color.r*factor);
	color.g += (int)(light->color.g*factor);
	color.b += (int)(light->color.b*factor);
	
	map_set(map_id, x, y, kColorLayer, (color_t)color);
}

void CalcLight(Light& light)
{
	fov_calc_cb(map_id, kOpacityLayer, light.x, light.y, light.radius, LightFOVCallback, &light);
}

std::vector<std::pair<int, int>> Closest9(int x, int y)
{
	std::vector<std::pair<int, int>> result;
	for (int dx=-1; dx<=+1; dx++)
	{
		for (int dy=-1; dy<=+1; dy++)
		{
			result.emplace_back(x+dx, y+dy);
		}
	}
	return result;
}

struct Point
{
	Point():
		x(),
		y()
	{ }
	
	Point(int x, int y):
		x(x),
		y(y)
	{ }
	
	bool operator== (const Point& rhs)
	{
		return x == rhs.x && y == rhs.y;
	}
	
	bool operator!= (const Point& rhs)
	{
		return x != rhs.x || y != rhs.y;
	}

	int x, y;
};

#include <iterator>
struct Range2D
{
	struct iterator: public std::iterator<std::forward_iterator_tag, Point>
	{
	public:
		typedef iterator self_type;
	
		iterator(int x, int y, int left, int right, int bottom):
			m_value(x, y),
			m_left(left),
			m_right(right),
			m_bottom(bottom)
		{ }
		
		self_type operator++()
		{
			self_type i = *this;
			if (m_value.x <= m_right)
			{
				m_value.x += 1;
				if (m_value.x > m_right)
				{
					if (m_value.y < m_bottom)
					{
						m_value.x = m_left;
						m_value.y += 1;
					}
				}
			}
			return i;
		}
		
		self_type operator++(int junk)
		{
			this->operator++();
			return *this;
		}
		
		reference operator*()
		{
			return m_value;
		}
		
		pointer operator->()
		{
			return &m_value;
		}
		
		bool operator==(const self_type& rhs)
		{
			return m_value == rhs.m_value;
		}
		
		bool operator!=(const self_type& rhs)
		{
			return m_value != rhs.m_value;
		}

	private:
		Point m_value;
		int m_left, m_right, m_bottom;
	};
	
	Range2D(int x, int y, int width, int height):
		m_x(x),
		m_y(y),
		m_width(width),
		m_height(height)
	{ }
	
	iterator begin()
	{
		return iterator(m_x, m_y, m_x, m_x+m_width-1, m_y+m_height-1);
	}
	
	iterator end()
	{
		return iterator(m_x+m_width, m_y+m_height-1, m_x, m_x+m_width-1, m_y+m_height-1);
	}
	
	int m_x, m_y, m_width, m_height;
};


Color3i AverageWallColor(int x, int y)
{
	int width = map_width(map_id);
	int height = map_height(map_id);
	Color3i result;
	int count = 0;
	
	for (auto& i: Range2D(x-1, y-1, 3, 3)) // 3x3 cells around (x, y)
	{
		if (i.x < 0 || i.x >= width || i.y < 0 || i.y >= height) continue; // OOB
		if (map_get(map_id, i.x, i.y, kFOVLayer) != 1) continue; // Must be in player's FOV
		if (map_get(map_id, i.x, i.y, kOpacityLayer) != 0) continue; // Must be a floor cell
		result += map_get(map_id, i.x, i.y, kColorLayer);
		count += 1;
	}
	
	if (count > 0)
	{
		result.r /= count;
		result.g /= count;
		result.b /= count;
	}
	
	return result;
}

void CalcWalls()
{
	for (int y=0; y<map_height(map_id); y++)
	{
		for (int x=0; x<map_width(map_id); x++)
		{
			if (map_get(map_id, x, y, kFOVLayer) != 1) continue; // Must be in player's FOV
			if (map_get(map_id, x, y, kOpacityLayer) != 1) continue; // Must be a wall
			map_set(map_id, x, y, kColorLayer, (color_t)AverageWallColor(x, y));
		}
	}
}

void DrawSimple()
{
	// Wall colors
	CalcWalls();
	
	// Cells
	for (int y=0; y<map_height(map_id); y++)
	{
		for (int x=0; x<map_width(map_id); x++)
		{	
			if (!map_get(map_id, x, y, kFOVLayer)) continue; // // Must be in player's FOV
			int c = map_get(map_id, x, y, kOpacityLayer)? 0x2588: 0xE000+32+4;
			terminal_color(map_get(map_id, x, y, kColorLayer));
			terminal_put(x, y, c);
		}
	}
}



std::vector<Color3i> CalcCornersColors()
{
	std::vector<Color3i> result;
	
	int width = map_width(map_id);
	int height = map_width(map_id);
	result.resize((width+1)*(height+1));
	
	for (int y=1; y<height; y++)
	{
		for (int x=1; x<width; x++)
		{
			Color3i color;
			int count = 0;
			
			// Accumulate
			for (auto& i: Range2D(x-1, y-1, 2, 2))
			{
				if (map_get(map_id, i.x, i.y, kOpacityLayer)) continue; // Must be a floor cell
				color += map_get(map_id, i.x, i.y, kColorLayer);
				count += 1;
			}
			
			// Normalize
			if (count > 0)
			{
				color.r /= count;
				color.g /= count;
				color.b /= count;
			}
			
			result[y*(width+1)+x] = color;
		}
	}
	
	return result;
}

void DrawAdvanced()
{
	int width = map_width(map_id);
	int height = map_width(map_id);
	std::vector<Color3i> corners = CalcCornersColors();

	terminal_blending(BLENDING_ADDITIVE);
	for (int y=0; y<height; y++)
	{
		for (int x=0; x<width; x++)
		{	
			if (!map_get(map_id, x, y, kFOVLayer)) continue; // Must be in player's FOV
			
			if (map_get(map_id, x, y, kOpacityLayer))
			{	
				// Left side
				if (x > 0 && !map_get(map_id, x-1, y, kOpacityLayer) && map_get(map_id, x-1, y, kFOVLayer))
				{
					int tile = 0;
					if (!map_get(map_id, x, y-1, kOpacityLayer)) tile |= 2;
					if (!map_get(map_id, x, y+1, kOpacityLayer)) tile |= 4;
				
					color_t color = corners[y*(width+1)+x];
					if (!map_get(map_id, x, y-1, kFOVLayer) || (map_get(map_id, x, y-1, kOpacityLayer) && !map_get(map_id, x-1, y-1, kFOVLayer))) color = 0xFF000000;
					terminal_color(color);
					terminal_put(x, y, 0xE000+tile+0);
					
					color = corners[(y+1)*(width+1)+x];
					if (!map_get(map_id, x, y+1, kFOVLayer) || (map_get(map_id, x, y+1, kOpacityLayer) && !map_get(map_id, x-1, y+1, kFOVLayer))) color = 0xFF000000;
					terminal_color(color);
					terminal_put(x, y, 0xE000+tile+1);
				}
				// Top side
				if (y > 0 && !map_get(map_id, x, y-1, kOpacityLayer) && map_get(map_id, x, y-1, kFOVLayer))
				{
					int tile = 8;
					if (!map_get(map_id, x-1, y, kOpacityLayer)) tile |= 2;
					if (!map_get(map_id, x+1, y, kOpacityLayer)) tile |= 4;
				
					color_t color = corners[y*(width+1)+x];
					if (!map_get(map_id, x-1, y, kFOVLayer) || (map_get(map_id, x-1, y, kOpacityLayer) && !map_get(map_id, x-1, y-1, kFOVLayer))) color = 0xFF000000;
					terminal_color(color);
					terminal_put(x, y, 0xE000+tile+0);
					
					color = corners[y*(width+1)+(x+1)];
					if (!map_get(map_id, x+1, y, kFOVLayer) || (map_get(map_id, x+1, y, kOpacityLayer) && !map_get(map_id, x+1, y-1, kFOVLayer))) color = 0xFF000000;
					terminal_color(color);
					terminal_put(x, y, 0xE000+tile+1);
				}
				// Right side
				if (x < width-1 && !map_get(map_id, x+1, y, kOpacityLayer) && map_get(map_id, x+1, y, kFOVLayer))
				{
					int tile = 16;
					if (!map_get(map_id, x, y-1, kOpacityLayer)) tile |= 2;
					if (!map_get(map_id, x, y+1, kOpacityLayer)) tile |= 4;
				
					color_t color = corners[y*(width+1)+(x+1)];
					if (!map_get(map_id, x, y-1, kFOVLayer) || (map_get(map_id, x, y-1, kOpacityLayer) && !map_get(map_id, x+1, y-1, kFOVLayer))) color = 0xFF000000;
					terminal_color(color);
					terminal_put(x, y, 0xE000+tile+0);
					
					color = corners[(y+1)*(width+1)+(x+1)];
					if (!map_get(map_id, x, y+1, kFOVLayer) || (map_get(map_id, x, y+1, kOpacityLayer) && !map_get(map_id, x+1, y+1, kFOVLayer))) color = 0xFF000000;
					terminal_color(color);
					terminal_put(x, y, 0xE000+tile+1);
				}
				// Bottom side
				if (y < height-1 && !map_get(map_id, x, y+1, kOpacityLayer) && map_get(map_id, x, y+1, kFOVLayer))
				{
					int tile = 24;
					if (!map_get(map_id, x-1, y, kOpacityLayer)) tile |= 2;
					if (!map_get(map_id, x+1, y, kOpacityLayer)) tile |= 4;
					
					color_t color = corners[(y+1)*(width+1)+x];
					if (!map_get(map_id, x-1, y, kFOVLayer) || (map_get(map_id, x-1, y, kOpacityLayer) && !map_get(map_id, x-1, y+1, kFOVLayer))) color = 0xFF000000;
					terminal_color(color);
					terminal_put(x, y, 0xE000+tile+0);
					
					color = corners[(y+1)*(width+1)+(x+1)];
					if (!map_get(map_id, x+1, y, kFOVLayer) || (map_get(map_id, x+1, y, kOpacityLayer) && !map_get(map_id, x+1, y+1, kFOVLayer))) color = 0xFF000000;
					terminal_color(color);
					terminal_put(x, y, 0xE000+tile+1);
				}
				
				// Top-left corner
				if (map_get(map_id, x, y-1, kFOVLayer) && map_get(map_id, x-1, y, kFOVLayer) && map_get(map_id, x-1, y-1, kFOVLayer) && map_get(map_id, x, y-1, kOpacityLayer) && map_get(map_id, x-1, y, kOpacityLayer))
				{
					terminal_color(corners[y*(width+1)+x]);
					terminal_put(x, y, 0xE000+32+0);
				}
				// Top-right corner
				if (map_get(map_id, x, y-1, kFOVLayer) && map_get(map_id, x+1, y, kFOVLayer) && map_get(map_id, x+1, y-1, kFOVLayer) && map_get(map_id, x, y-1, kOpacityLayer) && map_get(map_id, x+1, y, kOpacityLayer))
				{
					terminal_color(corners[y*(width+1)+(x+1)]);
					terminal_put(x, y, 0xE000+32+1);
				}
				// Bottom-left corner
				if (map_get(map_id, x, y+1, kFOVLayer) && map_get(map_id, x-1, y, kFOVLayer) && map_get(map_id, x-1, y+1, kFOVLayer) && map_get(map_id, x, y+1, kOpacityLayer) && map_get(map_id, x-1, y, kOpacityLayer))
				{
					terminal_color(corners[(y+1)*(width+1)+x]);
					terminal_put(x, y, 0xE000+32+2);
				}
				// Bottom-right corner
				if (map_get(map_id, x, y+1, kFOVLayer) && map_get(map_id, x+1, y, kFOVLayer) && map_get(map_id, x+1, y+1, kFOVLayer) && map_get(map_id, x, y+1, kOpacityLayer) && map_get(map_id, x+1, y, kOpacityLayer))
				{
					terminal_color(corners[(y+1)*(width+1)+(x+1)]);
					terminal_put(x, y, 0xE000+32+3);
				}
			}
			else
			{
				// Floor
				terminal_color(map_get(map_id, x, y, kColorLayer));
				terminal_put(x, y, 0xE000+32+4);
			}
		}
	}
	terminal_blending(BLENDING_NONE);
}

void FillAmbient()
{
	color_t ambient = color_from_name("darkest gray");
	for (int y=0; y<map_height(map_id); y++)
	{
		for (int x=0; x<map_width(map_id); x++)
		{
			map_set(map_id, x, y, kColorLayer, ambient);
		}
	}
}

void MovePlayer(int dx, int dy)
{
	int nx = player.x + dx;
	int ny = player.y + dy;
	
	if (nx < 0 || nx >= map_width(map_id) || ny < 0 || ny >= map_height(map_id))
	{
		return;
	}
	
	if (map_get(map_id, nx, ny, kOpacityLayer) > 0) // Wall
	{
		return;
	}
	
	player.x = nx;
	player.y = ny;
}

int Main()
{
	if (!LoadMap())
	{
		ErrorMessage("Couldn't load a map from \"map.txt\" file");
		return -1;
	}
	
	terminal_open(CODEPAGE_DEFAULT);
	terminal_options("size=%dx%d", map_width(map_id), map_height(map_id)+1);
	terminal_options("title=Lighting");
	terminal_options("font.name=media\\UbuntuMono-R.ttf; font.size=12");
	terminal_options("font.supplement=media\\supplement.png");
	
	while (true)
	{
		terminal_clear();
		
		FillAmbient();
	
		// Lights' and player's FOVs
		for (auto& i: lights) if (i.second.enabled) CalcLight(i.second);
		map_clear_layer(map_id, kFOVLayer);
		fov_calc(map_id, kOpacityLayer, kFOVLayer, lights[0].x, lights[0].y, player_sight_radius);
		
		// Render walls
		advanced? DrawAdvanced(): DrawSimple();
		
		// Lights' markers
		for (int id=9; id>=0; id--)
		{
			auto& light = lights[id];
			
			if (map_get(map_id, light.x, light.y, kFOVLayer))
			{
				terminal_color(light.enabled? (color_t)light.color: color_from_name("gray"));
				terminal_put(light.x, light.y, id? '0'+id: '@');
			}
		}
		
		// Some info
		terminal_color(color_from_name("white"));
		terminal_printf(0, map_height(map_id), "Lighting mode [[use TAB]]: %s", advanced? "advanced": "simple");
		
		terminal_refresh();
		
		int key = terminal_get();
		if (key == VK_CLOSE || key == VK_ESCAPE)
		{
			break;
		}
		else if (key >= VK_0 && key <= VK_9)
		{
			auto& light = lights[key-VK_0];
			light.enabled = !light.enabled;
		}
		else if (kMovement.count(key))
		{
			MovePlayer(kMovement[key].first, kMovement[key].second);
		}
		else if (key == VK_TAB)
		{
			advanced = !advanced;
		}
	}
	
	terminal_close();

	return 0;
}

#ifdef __linux
void ErrorMessage(const std::string& message)
{
	std::cerr << message << std::endl;
}

int main()
{
	return Main();
}
#else
#define WIN32_LEAN_AND_MEAN
#include <Windows.h>

void ErrorMessage(const std::string& message)
{
	MessageBoxA(NULL, message.c_str(), "Lighting", MB_OK);
}

int CALLBACK WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow)
{
	return Main();
}
#endif
