# Generic makefile

# Variables NAME, TYPE and DEPENDS must be defined by caller makefile

# Platform detection
PLATFORM = Unknown
ifeq ($(OS),Windows_NT)
    PLATFORM = Windows
else
    UNAME_S := $(shell uname -s)
    ifeq ($(UNAME_S),Linux)
    	PLATFORM = Linux
        CCFLAGS += -D LINUX
    endif
endif
ifeq ($(PLATFORM),Unknown)
	$(error "Unknown or unsupported platform")
endif


# Common settings
CXX = g++
CXXFLAGS = -Wall -std=c++11 -O3 -g0 -flto
LDFLAGS = -s -flto -fvisibility=hidden

# Platform-specific settings
ifeq ($(PLATFORM),Linux)
	LDFLAGS += -static-libstdc++ -static-libgcc
	SRCDIR = ./Source
	OUTDIR = ./Output
	OBJDIR = $(OUTDIR)/Obj
	# Linux .so dependencies must be linked via -Lpath -lname and not by path/filename
	# or else loader will search them in exact path regardless of -Wl,-R. option. 
	PROJECT_DEPS = $(filter-out header:% system:%,$(DEPENDS))
	LDFLAGS += $(foreach dep,$(PROJECT_DEPS),-L../$(dep)/Output)
	LIBS += $(foreach dep,$(PROJECT_DEPS),-l$(dep))
	ifeq ($(TYPE),library)
		ARTIFACT = lib$(NAME).so
		SATELLITES = $(SRCDIR)/$(NAME).h
		LDFLAGS += -shared
	endif
	ifeq ($(TYPE),$(filter $(TYPE),app cli))
		ARTIFACT = $(NAME)
		LDFLAGS += -Wl,-R.
		SATELLITES += $(foreach dep,$(PROJECT_DEPS),../$(dep)/Output/lib$(dep).so)
	endif
endif
ifeq ($(PLATFORM),Windows)
	LDFLAGS += -static
	SRCDIR = .\Source
	OUTDIR = .\Output
	OBJDIR = $(OUTDIR)\Obj
	# Windows .dll dependencies may be linked via path/filename since Windows
	# loader will search for them in CWD anyways.
	PROJECT_DEPS = $(filter-out header:% system:%,$(DEPENDS))
	LIBS += $(foreach dep,$(PROJECT_DEPS),..\$(dep)\Output\$(dep).dll)
	ifeq ($(TYPE),library)
		ARTIFACT = $(NAME).dll
		SATELLITES = $(SRCDIR)\$(NAME).h
		LDFLAGS += -shared -Wl,--out-implib,$(OUTDIR)\$(NAME).lib
	endif
	ifeq ($(TYPE),$(filter $(TYPE),app cli))
		ARTIFACT = $(NAME).exe
		SATELLITES += $(foreach dep,$(PROJECT_DEPS),..\$(dep)\Output\$(dep).dll)
	endif
	ifeq ($(TYPE),app)
		LDFLAGS += -mwindows
	endif
endif

# Include paths/libraries project depends on
CXXFLAGS += -I$(SRCDIR)
CXXFLAGS += $(foreach dep,$(filter-out system:%,$(DEPENDS)),-I../$(subst header:,,$(dep))/Output)
CXXFLAGS += $(foreach inc,$(ADDITIONAL_INCLUDES),-I$(inc))
CXXFLAGS += $(ADDITIONAL_CFLAGS)
LIBS += $(foreach dep,$(filter system:%,$(DEPENDS)),-l$(subst system:,,$(dep)))
LIBS += $(ADDITIONAL_LIBS)
LDFLAGS += $(ADDITIONAL_LDFLAGS)

# Static dependencies
CXXFLAGS += -I./Misc/include
LDFLAGS += -L./Misc/lib
# TODO: all lib*.a files in ./Misc/lib must be linked to project
# TODO: maybe leave them as-is, linking via ADDITIONAL_LIBS?

# Sources
CPPFILES = $(wildcard $(SRCDIR)/*.cpp)
OBJFILES = $(addprefix $(OBJDIR)/,$(notdir $(CPPFILES:.cpp=.o)))
DEPFILES = $(OBJFILES:.o=.d)


# Linux target implementation
ifeq ($(PLATFORM),Linux)

all: project

project: announce objdir artifact
	@if [ -f "$(OUTDIR)/$(ARTIFACT)" ]; then \
		echo "Build successful"; \
		for f in $(SATELLITES); do \
			cp -R "$$f" $(OUTDIR); \
		done; \
	fi

announce:
	@echo Building $(TYPE) $(NAME) on $(PLATFORM)

artifact: $(OBJFILES)
	@echo Linking $(ARTIFACT)
	@$(CXX) $(LDFLAGS) -o $(OUTDIR)/$(ARTIFACT) $^ $(LIBS)

$(OBJDIR)/%.o: $(SRCDIR)/%.cpp
	@echo Compiling $<
	@$(CXX) $(CXXFLAGS) -MMD -c -o $@ $<

outdir:
	@if [ ! -d "$(OUTDIR)" ]; then mkdir "$(OUTDIR)"; fi

objdir: outdir
	@if [ ! -d "$(OBJDIR)" ]; then mkdir "$(OBJDIR)"; fi

clean:
	@rm -rf $(OBJDIR)

endif


# Windows target implementation
ifeq ($(PLATFORM),Windows)

all: project

project: announce objdir artifact
	@if exist $(OUTDIR)\$(ARTIFACT) \
	( \
		(echo Build successful) && \
		(for %%I in ($(SATELLITES)) do (copy /y %%I $(OUTDIR) >nul 2>&1)) \
	)

announce:
	@echo Building $(TYPE) $(NAME) on $(PLATFORM)

artifact: $(OBJFILES)
	@echo Linking $(ARTIFACT)
	@$(CXX) $(LDFLAGS) -o $(OUTDIR)\$(ARTIFACT) $^ $(LIBS)

$(OBJDIR)/%.o: $(SRCDIR)/%.cpp
	@echo Compiling $<
	@$(CXX) $(CXXFLAGS) -MMD -c -o $@ $<

outdir:
	@if not exist $(OUTDIR)\nul (mkdir $(OUTDIR))

objdir: outdir
	@if not exist $(OBJDIR)\nul (mkdir $(OBJDIR))
	
clean:
	@if exist $(OBJDIR)\nul (@rmdir /s /q $(OBJDIR))

endif

# Automatic header dependencies
-include $(DEPFILES)

